def init():
    return {"running": 0, "pending": 0, "error": 0, "lost": 0}


def is_equal(a, b):
    for key in a:
        if a[key] != b[key]:
            return False
    return True


def estimate(
    num_jobs_running,
    num_jobs_pending,
    num_jobs_error,
    num_resubmissions_by_ichunk,
    max_num_resubmissions,
):
    num_jobs = init()
    num_jobs["running"] = num_jobs_running
    num_jobs["pending"] = num_jobs_pending
    num_jobs["error"] = num_jobs_error
    num_jobs["lost"] = 0
    for ichunk in num_resubmissions_by_ichunk:
        if num_resubmissions_by_ichunk[ichunk] >= max_num_resubmissions:
            num_jobs["lost"] += 1
    return num_jobs


def to_str(a):
    return "{: 4d} running, {: 4d} pending, {: 4d} error, {: 4d} lost".format(
        a["running"],
        a["pending"],
        a["error"],
        a["lost"],
    )



import json_line_logger


def filter_jobs_by_jobnames(jobs, jobnames):
    jobnames = set(jobnames)
    outjobs = []
    for job in jobs:
        if job["name"] in jobnames:
            outjobs.append(job)
    return outjobs


def split_jobs_in_running_pending_error(jobs, logger=None):
    if logger is None:
        logger = json_line_logger.LoggerStdout()

    running = []
    pending = []
    error = []

    for job in jobs:
        if job["state"] == "RUNNING":
            running.append(job)
        elif job["state"] == "PENDING":
            pending.append(job)
        elif "err" in str.lower(job["state"]):
            error.append(job)
        else:
            logger.degug(
                "job {:s} is in state {:s}.".format(job["name"], job["state"])
            )

    return running, pending, error



from . import call
from . import organizing_jobs
from .. import proto_pool
from .. import utils


def Pool(
    num_chunks=None,
    python_path=None,
    polling_interval=5.0,
    work_dir=None,
    keep_work_dir=False,
    max_num_resubmissions=0,
    # slurm specific
    # --------------
    sbatch_path="sbatch",
    clusters=None,
    squeue_path="squeue",
    scancel_path="scancel",
):
    return proto_pool.Pool(
        num_chunks=num_chunks,
        python_path=python_path,
        polling_interval=polling_interval,
        work_dir=work_dir,
        keep_work_dir=keep_work_dir,
        max_num_resubmissions=max_num_resubmissions,
        job_submission_function=submit,
        job_submission_function_kwargs={
            "sbatch_path": sbatch_path,
            "clusters": clusters,
        },
        job_query_status_function=status,
        job_query_status_function_kwargs={
            "squeue_path": squeue_path,
        },
        job_deletion_function=delete,
        job_deletion_function_kwargs={"scancel_path": scancel_path},
    )


def submit(
    jobname,
    script_path,
    script_arguments,
    stdout_path,
    stderr_path,
    logger,
    # slurm specific
    # --------------
    sbatch_path="sbatch",
    clusters=None,
):
    return call.sbatch(
        script_path=script_path,
        script_arguments=script_arguments,
        stdout_path=stdout_path,
        stderr_path=stderr_path,
        jobname=jobname,
        logger=logger,
        clusters=clusters,
        sbatch_path=sbatch_path,
    )


def status(
    jobnames,
    logger,
    # slurm specific
    # --------------
    squeue_path,
):
    all_jobs = call.squeue(
        squeue_path=squeue_path,
        logger=logger,
    )
    our_jobs = organizing_jobs.filter_jobs_by_jobnames(
        jobs=all_jobs,
        jobnames=jobnames,
    )
    (
        running,
        pending,
        error,
    ) = organizing_jobs.split_jobs_in_running_pending_error(
        jobs=our_jobs,
        logger=logger,
    )
    out = {
        "running": [],
        "pending": [],
        "error": [],
    }
    for job in running:
        out["running"].append(_make_job(slurm_job=job))
    for job in pending:
        out["pending"].append(_make_job(slurm_job=job))
    for job in error:
        out["error"].append(_make_job(slurm_job=job))
    return out


def delete(
    job,
    logger,
    # slurm specific
    # --------------
    scancel_path,
):
    return call.scancel(
        jobname=job["name"],
        scancel_path=scancel_path,
        logger=logger,
    )


def _make_job(slurm_job):
    return {
        "name": slurm_job["name"],
    }



"""
A dummy queue for testing qsub, qstat, and qdel.
"""
import os
import json
from .. import utils


def dummy_paths():
    join = os.path.join
    pypoolparty_dir = utils.resources_path()
    ddir = join(pypoolparty_dir, "sun_grid_engine", "tests", "resources")
    out = {}
    out["queue_state"] = join(ddir, "dummy_queue_state.json")
    out["qsub"] = join(ddir, "dummy_qsub.py")
    out["qstat"] = join(ddir, "dummy_qstat.py")
    out["qdel"] = join(ddir, "dummy_qdel.py")
    return out


def init_queue_state(path, evil_jobs=[]):
    with open(path, "wt") as f:
        f.write(
            json.dumps({"running": [], "pending": [], "evil_jobs": evil_jobs})
        )



import math


def assign_tasks_to_chunks(num_tasks, num_chunks):
    """
    When you have too many tasks for your parallel processing queue this
    function chunks multiple tasks into fewer chunks.

    Parameters
    ----------
    num_tasks : int
        Number of tasks.
    num_chunks : int (optional)
        The maximum number of chunks. Your tasks will be spread over
        these many chunks. If None, each chunk contains a single task.

    Returns
    -------
        A list of chunks where each chunk is a list of task-indices `itask`.
        The lengths of the list of chunks is <= num_chunks.
    """
    if num_chunks is None:
        num_tasks_in_chunk = 1
    else:
        assert num_chunks > 0
        num_tasks_in_chunk = int(math.ceil(num_tasks / num_chunks))

    chunks = []
    current_chunk = []
    for j in range(num_tasks):
        if len(current_chunk) < num_tasks_in_chunk:
            current_chunk.append(j)
        else:
            chunks.append(current_chunk)
            current_chunk = []
            current_chunk.append(j)
    if len(current_chunk):
        chunks.append(current_chunk)
    return chunks



from . import testing
from . import call
from . import organizing_jobs
from .. import proto_pool
from .. import utils


def Pool(
    num_chunks=None,
    python_path=None,
    polling_interval=5.0,
    work_dir=None,
    keep_work_dir=False,
    max_num_resubmissions=0,
    # sge specific
    # ------------
    qsub_path="qsub",
    queue_name=None,
    qstat_path="qstat",
    error_state_indicator="E",
    qdel_path="qdel",
):
    if python_path is None:
        python_path = utils.default_python_path()

    return proto_pool.Pool(
        num_chunks=num_chunks,
        python_path=python_path,
        polling_interval=polling_interval,
        work_dir=work_dir,
        keep_work_dir=keep_work_dir,
        max_num_resubmissions=max_num_resubmissions,
        job_submission_function=submit,
        job_submission_function_kwargs={
            "qsub_path": qsub_path,
            "queue_name": queue_name,
            "script_exe_path": python_path,
        },
        job_query_status_function=status,
        job_query_status_function_kwargs={
            "qstat_path": qstat_path,
            "error_state_indicator": error_state_indicator,
        },
        job_deletion_function=delete,
        job_deletion_function_kwargs={"qdel_path": qdel_path},
    )


def submit(
    jobname,
    script_path,
    script_arguments,
    stdout_path,
    stderr_path,
    logger,
    # sge specific
    # ------------
    qsub_path,
    queue_name,
    script_exe_path,
):
    return call.qsub(
        qsub_path=qsub_path,
        queue_name=queue_name,
        script_exe_path=script_exe_path,
        script_path=script_path,
        arguments=script_arguments,
        JB_name=jobname,
        stdout_path=stdout_path,
        stderr_path=stderr_path,
        logger=logger,
    )


def status(
    jobnames,
    logger,
    # sge specific
    # ------------
    qstat_path,
    error_state_indicator,
):
    all_jobs_running, all_jobs_pending = call.qstat(
        qstat_path=qstat_path,
        logger=logger,
    )
    running, pending, error = organizing_jobs.get_jobs_running_pending_error(
        JB_names_set=jobnames,
        all_jobs_running=all_jobs_running,
        all_jobs_pending=all_jobs_pending,
        error_state_indicator=error_state_indicator,
    )
    out = {
        "running": [],
        "pending": [],
        "error": [],
    }
    for job in running:
        out["running"].append(_make_job(sge_job=job))
    for job in pending:
        out["pending"].append(_make_job(sge_job=job))
    for job in error:
        out["error"].append(_make_job(sge_job=job))
    return out


def delete(
    job,
    logger,
    # sge specific
    # ------------
    qdel_path,
):
    return call.qdel(
        JB_job_number=job["JB_job_number"],
        qdel_path=qdel_path,
        logger=logger,
    )


def _make_job(sge_job):
    return {
        "name": sge_job["JB_name"],
        "JB_job_number": sge_job["JB_job_number"],
    }



import os
import stat
import shutil
import time
import rename_after_writing
import pickle


def make_path_executable(path):
    st = os.stat(path)
    os.chmod(path, st.st_mode | stat.S_IEXEC)


def default_python_path():
    return os.path.abspath(shutil.which("python"))


def session_id_from_time_now():
    # This must be a valid filename. No ':' for time.
    return time.strftime("%Y-%m-%dT%H-%M-%S", time.gmtime())


def read(path, mode="t"):
    with open(path, mode + "r") as f:
        content = f.read()
    return content


def write(path, content, mode="t"):
    with rename_after_writing.open(file=path, mode=mode + "w") as f:
        f.write(content)


def read_text(path):
    return read(path=path, mode="t")


def write_text(path, content):
    write(path=path, content=content, mode="t")


def read_pickle(path):
    return pickle.loads(read(path=path, mode="b"))


def write_pickle(path, content):
    write(path=path, content=pickle.dumps(content), mode="b")


def resources_path(package_name="pypoolparty"):
    try:
        # python version after 3.7
        import importlib
        from importlib import resources

        return str(importlib.resources.files(package_name))
    except Exception as err:
        pass

    # python version up to 3.7
    import pkg_resources

    return str(
        pkg_resources.resource_filename(
            package_or_requirement=package_name,
            resource_name="",
        )
    )


def shutdown_logger(logger):
    for fh in logger.handlers:
        fh.flush()
        fh.close()
        logger.removeHandler(fh)



import os
import stat
import time
from . import utils


def session_id_from_time_now():
    # This must be a valid filename. No ':' for time.
    return time.strftime("%Y-%m-%dT%H-%M-%S", time.gmtime())


def make_jobname_from_ichunk(session_id, ichunk):
    return "q{:s}#{:09d}".format(session_id, ichunk)


def make_ichunk_from_jobname(jobname):
    ichunk_str = jobname.split("#")[1]
    return int(ichunk_str)


def chunk_path(work_dir, ichunk):
    return os.path.join(work_dir, "{:09d}.pkl".format(ichunk))


def map_tasks_into_work_dir(work_dir, tasks, chunks, session_id):
    jobnames_in_session = set()
    for ichunk, chunk in enumerate(chunks):
        jobname = make_jobname_from_ichunk(
            session_id=session_id,
            ichunk=ichunk,
        )
        jobnames_in_session.add(jobname)
        chunk_payload = [tasks[itask] for itask in chunk]
        utils.write_pickle(
            path=chunk_path(work_dir, ichunk),
            content=chunk_payload,
        )
    return jobnames_in_session


def reduce_task_results_from_work_dir(work_dir, chunks, logger):
    task_results = []
    task_results_are_incomplete = False

    for ichunk, chunk in enumerate(chunks):
        num_tasks_in_chunk = len(chunk)
        chunk_result_path = chunk_path(work_dir, ichunk) + ".out"

        try:
            chunk_result = utils.read_pickle(path=chunk_result_path)
            for task_result in chunk_result:
                task_results.append(task_result)
        except FileNotFoundError:
            task_results_are_incomplete = True
            logger.warning(
                "Expected results in: {:s}".format(chunk_result_path)
            )
            task_results += [None for i in range(num_tasks_in_chunk)]

    return task_results_are_incomplete, task_results


def has_invalid_or_non_empty_stderr(work_dir, num_chunks):
    has_errors = False
    for ichunk in range(num_chunks):
        e_path = chunk_path(work_dir, ichunk) + ".e"
        try:
            if os.stat(e_path).st_size != 0:
                has_errors = True
        except FileNotFoundError:
            has_errors = True
    return has_errors



from . import utils
from . import making_script
from . import job_counter
from . import pooling
from . import chunking

import json_line_logger
import os
import shutil
import time
import json


class Pool:
    """
    Multiprocessing on a compute-cluster using queues.
    """

    def __init__(
        self,
        num_chunks=None,
        python_path=None,
        polling_interval=5.0,
        work_dir=None,
        keep_work_dir=False,
        max_num_resubmissions=10,
        job_submission_function=None,
        job_submission_function_kwargs=None,
        job_query_status_function=None,
        job_query_status_function_kwargs=None,
        job_deletion_function=None,
        job_deletion_function_kwargs=None,
    ):
        """
        Parameters
        ----------
        clusters : (list of str) or None
            Name of the clusters to submit to.
        python_path : str or None
            The python path to be used on the computing-cluster's worker-nodes
            to execute the worker-node's python-script.
        polling_interval : float or None
            The time in seconds to wait before polling squeue again while
            waiting for the jobs to finish.
        work_dir : str
            The directory path where the tasks, the results and the
            worker-node-script is stored.
        keep_work_dir : bool
            When True, the working directory will not be removed.
        max_num_resubmissions: int
            In case of error-state in queue-job, the job will be tried this
            often to be resubmitted befor giving up on it.
        num_chunks : int or None
            If provided, the tasks are grouped in this many chunks.
            The tasks in a chunk are computed in serial on the worker-node.
            It is useful to chunk tasks when the number of tasks is much larger
            than the number of available slots for parallel computing and the
            start-up-time for a slot is not much smaller than the compute-time
            for a single task.
        """
        if python_path is None:
            self.python_path = utils.default_python_path()
        else:
            self.python_path = python_path
        self.polling_interval = polling_interval
        self.work_dir = work_dir
        self.keep_work_dir = keep_work_dir
        self.max_num_resubmissions = max_num_resubmissions
        self.num_chunks = num_chunks

        self.submit_jobs = job_submission_function
        self.submit_jobs_kwargs = job_submission_function_kwargs
        self.delete_jobs = job_deletion_function
        self.delete_jobs_kwargs = job_deletion_function_kwargs
        self.status_jobs = job_query_status_function
        self.status_jobs_kwargs = job_query_status_function_kwargs

    def __repr__(self):
        return self.__class__.__name__ + "()"

    def map(self, func, iterable):
        """
        Apply `func` to each element in `iterable`, collecting the results
        in a list that is returned.

        Parameters
        ----------
        func : function-pointer
            Pointer to a function in a python-module. It must have both:
            func.__module__
            func.__name__
        iterable : list
            List of tasks. Each task must be a valid input to 'func'.

        Returns
        -------
        results : list
            Results. One result for each task.

        Example
        -------
        results = pool.map(sum, [[1, 2], [2, 3], [4, 5], ])
            [3, 5, 9]
        """
        tasks = iterable
        session_id = utils.session_id_from_time_now()

        if self.work_dir is None:
            swd = os.path.abspath(
                os.path.join(".", ".pypoolparty_" + session_id)
            )
        else:
            swd = os.path.abspath(self.work_dir)

        os.makedirs(swd)
        sl = json_line_logger.LoggerFile(path=os.path.join(swd, "log.jsonl"))

        sl.debug("Starting map()")
        sl.debug("python path: {:s}".format(self.python_path))
        sl.debug("polling-interval: {:f}s".format(self.polling_interval))
        sl.debug(
            "max. num. resubmissions: {:d}".format(self.max_num_resubmissions)
        )

        script_path = os.path.join(swd, "worker_node_script.py")
        sl.debug("Writing worker-node-script: {:s}".format(script_path))
        shebang = "#!{:s}".format(self.python_path)
        script_content = making_script.make(
            func_module=func.__module__,
            func_name=func.__name__,
            environ=dict(os.environ),
            shebang=shebang,
        )
        utils.write_text(path=script_path, content=script_content)
        utils.make_path_executable(path=script_path)

        sl.debug("Make chunks of tasks")

        chunks = chunking.assign_tasks_to_chunks(
            num_tasks=len(tasks),
            num_chunks=self.num_chunks,
        )

        sl.debug("Mapping chunks of tasks into work_dir")

        jobnames_in_session = pooling.map_tasks_into_work_dir(
            work_dir=swd,
            tasks=tasks,
            chunks=chunks,
            session_id=session_id,
        )

        sl.debug("Submitting jobs")

        for jobname in jobnames_in_session:
            ichunk = pooling.make_ichunk_from_jobname(jobname=jobname)
            self.submit_jobs(
                jobname=jobname,
                script_path=script_path,
                script_arguments=[pooling.chunk_path(swd, ichunk)],
                stdout_path=pooling.chunk_path(swd, ichunk) + ".o",
                stderr_path=pooling.chunk_path(swd, ichunk) + ".e",
                logger=sl,
                **self.submit_jobs_kwargs,
            )

        sl.debug("Waiting for jobs to finish")

        still_running = True
        num_resubmissions_by_ichunk = {}
        last_job_count = job_counter.init()

        while still_running:
            job_stati = self.status_jobs(
                jobnames=jobnames_in_session,
                logger=sl,
                **self.status_jobs_kwargs,
            )
            job_count = job_counter.estimate(
                num_jobs_running=len(job_stati["running"]),
                num_jobs_pending=len(job_stati["pending"]),
                num_jobs_error=len(job_stati["error"]),
                num_resubmissions_by_ichunk=num_resubmissions_by_ichunk,
                max_num_resubmissions=self.max_num_resubmissions,
            )

            if not job_counter.is_equal(job_count, last_job_count):
                sl.info(job_counter.to_str(job_count))
            last_job_count = job_count

            for job in job_stati["error"]:
                ichunk = pooling.make_ichunk_from_jobname(jobname=job["name"])
                if ichunk in num_resubmissions_by_ichunk:
                    num_resubmissions_by_ichunk[ichunk] += 1
                else:
                    num_resubmissions_by_ichunk[ichunk] = 1

                job_id_str = "name {:s}, ichunk {:09d}".format(
                    job["name"], ichunk
                )
                sl.warning("Found error-state in: {:s}".format(job_id_str))
                sl.warning("Deleting: {:s}".format(job_id_str))

                self.delete_jobs(job=job, logger=sl, **self.delete_jobs_kwargs)

                if (
                    num_resubmissions_by_ichunk[ichunk]
                    <= self.max_num_resubmissions
                ):
                    sl.warning(
                        "Resubmitting {:d} of {:d}, jobname {:s}".format(
                            num_resubmissions_by_ichunk[ichunk],
                            self.max_num_resubmissions,
                            job["name"],
                        )
                    )
                    self.submit_jobs(
                        jobname=job["name"],
                        script_path=script_path,
                        script_arguments=[pooling.chunk_path(swd, ichunk)],
                        stdout_path=pooling.chunk_path(swd, ichunk) + ".o",
                        stderr_path=pooling.chunk_path(swd, ichunk) + ".e",
                        logger=sl,
                        **self.submit_jobs_kwargs,
                    )

            if job_stati["error"]:
                utils.write_text(
                    path=os.path.join(swd, "num_resubmissions_by_ichunk.json"),
                    content=json.dumps(num_resubmissions_by_ichunk, indent=4),
                )

            if job_count["running"] == 0 and job_count["pending"] == 0:
                still_running = False

            time.sleep(self.polling_interval)

        sl.debug("Reducing results from work_dir")
        (
            task_results_are_incomplete,
            task_results,
        ) = pooling.reduce_task_results_from_work_dir(
            work_dir=swd,
            chunks=chunks,
            logger=sl,
        )

        has_stderr = pooling.has_invalid_or_non_empty_stderr(
            work_dir=swd, num_chunks=len(chunks)
        )
        if has_stderr:
            sl.warning(
                "At least one task wrote to std-error or was not processed at all"
            )

        if has_stderr or self.keep_work_dir or task_results_are_incomplete:
            remove_work_dir = False
            sl.warning("Keeping work_dir: {:s}".format(swd))
        else:
            remove_work_dir = True
            sl.debug("Removing work_dir: {:s}".format(swd))

        utils.shutdown_logger(logger=sl)
        del sl

        if remove_work_dir:
            shutil.rmtree(swd)

        return task_results



import setuptools
import os


with open("README.rst", "r", encoding="utf-8") as f:
    long_description = f.read()

with open(os.path.join("pypoolparty", "version.py")) as f:
    txt = f.read()
    last_line = txt.splitlines()[-1]
    version_string = last_line.split()[-1]
    version = version_string.strip("\"'")

setuptools.setup(
    name="pypoolparty",
    version=version,
    description=(
        "A job pool for distributed compute clusters inspired by "
        "python's multoprocessing.Pool."
    ),
    long_description=long_description,
    long_description_content_type="text/x-rst",
    url="https://github.com/cherenkov-plenoscope/pypoolparty",
    author="Sebastian Achim Mueller",
    author_email="sebastian-achim.mueller@mpi-hd.mpg.de",
    packages=[
        "pypoolparty",
        "pypoolparty.sun_grid_engine",
        "pypoolparty.slurm",
    ],
    package_data={
        "pypoolparty": [
            os.path.join("slurm", "tests", "resources", "*"),
            os.path.join("sun_grid_engine", "tests", "resources", "*"),
        ]
    },
    install_requires=[
        "qstat>=0.0.5",
        "json_line_logger",
        "rename_after_writing",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
        "Natural Language :: English",
    ],
)
